#!/usr/bin/python
# -*- coding: utf-8 -*-
#

#Copyright (C) 2006-2007  Néstor Arocha Rodríguez, Inmaculada Luengo Merino

#This file is part of Driza.
#
#Driza is free software; you can redistribute it and/or modify
#it under the terms of the GNU General Public License as published by
#the Free Software Foundation; either version 2 of the License, or
#(at your option) any later version.
#
#Driza is distributed in the hope that it will be useful,
#but WITHOUT ANY WARRANTY; without even the implied warranty of
#MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#GNU General Public License for more details.
#
#You should have received a copy of the GNU General Public License
#along with Driza; if not, write to the Free Software
#Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

"""
Datos: Todas las clases relacionadas con los datos, 
y sus correspondientes metodos
"""

import logging
log = logging.getLogger("Driza.datos.datos")

class Registro(list):
    """Almacena un registro, permitiendo el acceso tanto por el indice, 
    como por el nombre de variable.  Esta clase garantiza que la lista de 
    valores se ajusta a los tipos de la lista de variables
    Hereda de list
    """

    def __init__(self, lvariables):
        """ 
        Clase constructora
        lvariables es el array de variables al que esta asociado el registro
        Añade los valores por defecto para todos los atributos
        """
        #VARIABLES PRIVADAS
        self.__variables = lvariables
        list.__init__(self)
        self.rellenar()
    
    #FUNCIONES PUBLICAS
    def __getitem__(self, indice):
        """ Redefinicion del [] de la lista. 
        Si nos pasan una ristra, comprobamos si corresponde a alguna variable
        Si nos pasan un objeto, comprobamos si se corresponde a alguna variable
        Si nos pasan un entero, devolvemos el valor de esa posicion
        """
        miindice = self.__variables.obtenerindice(indice)
        return list.__getitem__(self, miindice)

    def __setitem__(self, indice, valor):
        """ 
        Redefinicion de []= de la lista. Almacena en el valor 
        convertido a formato interno por la variable asociada
        """
        miindice = self.__variables.obtenerindice(indice)
        list.__setitem__(self, miindice, self.__variables[miindice].nuevoitem\
                (valor))

    def rellenar(self):
        """Rellena el registro con los valores por defecto de las variables que faltan"""
        listadefault=self.__variables.arr_defaults()
        for i in range (self.__len__(),self.__variables.__len__()):
            intermedio = self.__variables[i].nuevoitem(listadefault[i])
            list.append(self, intermedio)

    def append(self, valor):
        """ 
        Redefinicion del append de lista. Añade al registro el valor pasado 
        en el formato interno de la variable asociada 
        """
        #TODO: Este append es problemático, puesto que depende de que 
        #la variable ya exista. Pendiente de rediseñar
        log.debug("Solicitado append de registro")
        assert(self.__len__()<len(self.__variables))
        intermedio = self.__variables[self.__len__()].nuevoitem(valor)
        list.append(self, intermedio)
        
    def set(self, lista):
        """ 
        Esta funcion establece los valores del registro, 
        teniendo como argumento una lista o derivado 
        """
        assert(isinstance(lista,list))
        i = 0
        for valor in lista:
            self.__setitem__(i, valor)
            i += 1


class ListaVar(list): 
    """
    ListaVar hereda de list. Su utilidad es almacenar un conjunto de 
    variables, permitiendo recuperarlas no solo por el indice, sino por 
    el nombre.
    """
    def __getcopy__(self):
        resultado = ListaVar()
        for i in self:
            resultado.append(i)
        return resultado

    #FUNCIONES PUBLICAS

    def arr_defaults(self):
        """
        Devuelve una lista con las ristras de todos 
        los valores por defecto
        """
        return [x.valorpordefecto for x in self]

    def obtenerindice(self, parametro):
        """
         Permite obtener el indice de una variable a partir de su nombre, 
         del objeto, o de un entero.  Devuelve un entero.
        """
        import types
        if parametro.__class__ == types.StringType:
            for iter in self:  
                if iter.nombre == parametro:
                    return self.index(iter)
        elif parametro.__class__ == types.IntType:
            return parametro
        elif parametro in self:
            return self.index(parametro)
        raise TypeError



class ContenedorRegVar:
    """
    Esta clase almacena los datos del programa. Tiene dos miembros, 
    una lista de registros y una lista de variables
    """
    def __init__(self, registros = None, variables = None):
        """Inicialización del contenedor principal"""
        if not registros: 
            registros=[]
        self.__reg = registros #Un array para datos
        if not variables: 
            variables = ListaVar()
        self.__var = variables #Un array de variables 

    def __deepcopy__(self, memo):
        """Reimplementación de __deepcopy__, permite una copia del objeto"""
        nuevalista = self.__var.__getcopy__()
        nuevalistaregistros = list()
        for registro in self.__reg:
            nuevoregistro = Registro(nuevalista)
            for x in range(len(registro)):
                nuevoregistro[x] = nuevalista[x].nuevoitem(repr(registro[x]))
            nuevalistaregistros.append(nuevoregistro)
        resultado= ContenedorRegVar(nuevalistaregistros, nuevalista)
        memo[id(self)] = resultado
        return resultado

    def registros(self):
        return self.__reg
    def variables(self):
        return self.__var




class ListaUndoRedo(list):
    """ListaUndoRedo hereda de lista. Es similar a una cola lifo.
    Los elementos almacenados son tuplas (estado,flagoriginal)
    """
    def __init__(self, tama):
        self.tama = tama #Almacena el tamaño almacenado

    def encolar(self, objeto, flagoriginal = False, index = 0):
        """Si pasan un indice es que quieren encolar desde alguna posición"""
        if index>=self.tama:
            raise IndexError
        else:
            if self.__len__() == self.tama:
                self.__delitem__(-1)
            self.__delslice__(0, index)
            self.insert(0, (objeto,flagoriginal))

    def setFlag(self, indice,valor):
        """Modifica la flag de original de un indice dado"""
        if (valor!=True and valor!=False):
            raise TypeError
        estadoactual=self.__getitem__(indice)[0]
        self.__setitem__(indice,(estadoactual,valor))

class PorteroDatos:
    """
    Clase de portera. Gestiona una lista con todas las copias de los datos, 
    determinando cual es la vigente y permitiendo el undo y el redo
    """
    def __init__(self,config):
        self.__index=0 #Variable que lleva el indice
        self.__lista = ListaUndoRedo(config.configuracion["nundo"])
        self.__lista.encolar(ContenedorRegVar(), True)

    def undo(self):
        """Realiza un undo, desplazando el indice de ListaUndoRedo"""
        log.debug("Solicitud de undo a portero")
        if self.__index < self.__lista.tama:
            log.debug("Undo ha resuelto satisfactoriamente")
            self.__index += 1
        else:
            raise IndexError

    def redo(self):
        """Realiza un redo, desplazando el indice de ListaUndoRedo"""
        if self.__index > 0:
            self.__index -= 1
        else:
            raise IndexError

    def puedoundo(self):
        """Devuelve un valor lógico indicando si es posible realizar un undo"""
        return ((self.__index<self.__lista.tama) and (len(self.__lista)>1) and \
                (self.__index<(len(self.__lista)-1)))

    def puedoredo(self):
        """Devuelve un valor lógico indicando si es posible realizar un redo"""
        return self.__index > 0

    def guardarestado(self, flagoriginal = False):
        """
        guarda un estado en la lista de estados, 
        creando una copia a partir del último
        La flagoriginal se activa si se esta abriendo un nuevo fichero
        """
        tmp = self.__lista[self.__index][0]
        import copy
        objeto= copy.deepcopy(tmp)
        self.__lista.encolar(objeto, flagoriginal, self.__index)
        self.__index=0
    
    def insertarestado(self, estado, flagoriginal = False):
        """Introduce un estado y lo encola"""
        self.__lista.encolar(estado, flagoriginal, self.__index)
        self.__index=0

    def nuevoestado(self, flagoriginal = False):
        """Introduce un estado nuevo y vacio"""
        self.__lista.encolar(ContenedorRegVar(), flagoriginal)
        self.__index=0

    def actual(self):
        """Devuelve el estado vigente"""
        return self.__lista[self.__index][0]

    def actualoriginal(self):
        """Devuelve si el estado vigente es original"""
        return self.__lista[self.__index][1]

    def setactualoriginal(self):
        """Establece que el estado actual es original"""
        self.__lista.setFlag(self.__index,True)

